import statistics
from rma.redis import *
from rma.helpers import *
from redis.exceptions import RedisError


class RealStringEntry:
    def get_int_encoded_bytes(self, redis, key_name):
        try:
            num_value = int(redis.get(key_name))
            if num_value < REDIS_SHARED_INTEGERS:
                return 0
            else:
                return size_of_pointer_fn()
        except ValueError:
            pass

        return size_of_pointer_fn()

    def __init__(self, key_name, redis):
        """
        :param key_name:
        :param RmaRedis redis:
        :return:
        """

        self.logger = logging.getLogger(__name__)
        try:
            self.encoding = redis.object("ENCODING", key_name).decode('utf8')
        except AttributeError as e:
            self.logger.warning("Invalid encoding from server for key `%s` (would be skipped)" % key_name)
            self.encoding = REDIS_ENCODING_EMBSTR
            self.useful_bytes = 0
            self.free_bytes = 0
            self.aligned = 0
            return

        if self.encoding == REDIS_ENCODING_INT:
            self.useful_bytes = self.get_int_encoded_bytes(redis, key_name)
            self.free_bytes = 0
            self.aligned = size_of_aligned_string_by_size(self.useful_bytes, encoding=self.encoding)
        elif self.encoding == REDIS_ENCODING_EMBSTR or self.encoding == REDIS_ENCODING_RAW:
            sdslen_response = redis.debug_sdslen(key_name)
            self.useful_bytes = sdslen_response['val_sds_len']
            self.free_bytes = sdslen_response['val_sds_avail']
            # INFO Rewrite this to support Redis >= 3.2 sds dynamic header
            sds_len = 8 + self.useful_bytes + self.free_bytes + 1
            self.aligned = size_of_aligned_string_by_size(sds_len, encoding=self.encoding)

    def __enter__(self):
        return self

    def __exit__(self, *exc):
        return False


class ValueString:
    def __init__(self, redis):
        self.redis = redis
        self.logger = logging.getLogger(__name__)

    def analyze(self, keys):
        key_stat = {
            'headers': ['Match', "Count", "Useful", "Free", "Real", "Ratio", "Encoding", "Min", "Max", "Avg"],
            'data': []
        }

        for pattern, data in keys.items():
            used_bytes = []
            free_bytes = []
            aligned_bytes = []
            encodings = []

            for key_name in data:
                try:
                    with RealStringEntry(key_name=key_name, redis=self.redis) as stat:
                        used_bytes.append(stat.useful_bytes)
                        free_bytes.append(stat.free_bytes)
                        aligned_bytes.append(stat.aligned)
                        encodings.append(stat.encoding)
                except RedisError as e:
                    # This code works in real time so key me be deleted and this code fail
                    self.logger.warning(repr(e))

            total_elements = len(used_bytes)
            used_user = sum(used_bytes)
            free_user = sum(free_bytes)
            aligned = sum(aligned_bytes)
            prefered_encoding = pref_encoding(encodings)

            min_bytes = min(used_bytes)
            mean = statistics.mean(used_bytes) if total_elements > 1 else min_bytes

            stat_entry = [
                pattern,
                total_elements,
                used_user,
                free_user,
                aligned,
                aligned / (used_user if used_user > 0 else 1),
                prefered_encoding,
                min_bytes,
                max(used_bytes),
                mean,
            ]
            key_stat['data'].append(stat_entry)

        key_stat['data'].sort(key=lambda e: e[1], reverse=True)
        key_stat['data'].append(make_total_row(key_stat['data'], ['Total:', sum, sum, 0, sum, 0, '', 0, 0, 0]))

        return [
            "String value stat",
            key_stat
        ]



#!/usr/bin/env python

import logging
from optparse import OptionParser
from rma.application import RmaApplication

logging.basicConfig(level=logging.INFO)

VALID_TYPES = ("string", "hash", "list", "set", "zset")
VALID_MODES = ('all', 'scanner', 'ram', 'global')


def main():
    usage = """usage: %prog [options]

Example : %prog -m * --type hash"""

    parser = OptionParser(usage=usage)
    parser.add_option("-s", "--server", dest="host", default="127.0.0.1",
                      help="Redis Server hostname. Defaults to 127.0.0.1")
    parser.add_option("-p", "--port", dest="port", default=6379, type="int",
                      help="Redis Server port. Defaults to 6379")
    parser.add_option("-a", "--password", dest="password",
                      help="Password to use when connecting to the server")
    parser.add_option("-d", "--db", dest="db", default=0,
                      help="Database number, defaults to 0")
    parser.add_option("-m", "--match", dest="match", default="*",
                      help="Keys pattern to match")
    parser.add_option("-l", "--limit", dest="limit", default="0", type="int",
                      help="Get max key matched by pattern")
    parser.add_option("-b", "--behaviour", dest="behaviour", default="all",
                      help="Specify application working mode")
    parser.add_option("-t", "--type", dest="types", action="append",
                      help="""Data types to include. Possible values are string, hash, list, set. Multiple types can be provided.
                    If not specified, all data types will be returned""")

    (options, args) = parser.parse_args()

    filters = {}
    if options.match:
        filters['match'] = options.match

    if options.behaviour:
        if options.behaviour not in VALID_MODES:
            raise Exception(
                'Invalid behaviour provided - %s. Expected one of %s' % (options.behaviour, (", ".join(VALID_TYPES))))
        else:
            filters['behaviour'] = options.behaviour

    if options.types:
        filters['types'] = []
        for x in options.types:
            if x not in VALID_TYPES:
                raise Exception('Invalid type provided - %s. Expected one of %s' % (x, (", ".join(VALID_TYPES))))
            else:
                filters['types'].append(x)

    app = RmaApplication(host=options.host, port=options.port, db=options.db,
                         password=options.password, match=options.match, limit=options.limit, filters=filters)
    app.run()


if __name__ == '__main__':
    main()



from rma.redis import *
from itertools import tee
from rma.helpers import *

import statistics


class ListStatEntry:
    def __init__(self, key_name, redis):
        """
        :param key_name:
        :param RmaRedis redis:
        :return:
        """
        self.values = redis.lrange(key_name, 0, -1)
        self.encoding = redis.object('encoding', key_name).decode('utf8')
        self.count = len(self.values)

        m, m2, m3 = tee((len(x) for x in self.values), 3)

        if self.encoding == 'linkedlist':
            self.system = dict_overhead(self.count)
            self.valueAlignedBytes = sum(map(size_of_linkedlist_aligned_string, self.values))
        elif self.encoding == 'ziplist' or self.encoding == 'quicklist':
            # Undone `quicklist`
            self.system = ziplist_overhead(self.count)
            self.valueAlignedBytes = sum(map(size_of_ziplist_aligned_string, self.values))
        else:
            raise Exception('Panic', 'Unknown encoding %s in %s' % (self.encoding, key_name))

        self.valueUsedBytes = sum(m)
        self.valueMin = min(m2)
        self.valueMax = max(m3)


class ListAggreegator:
    def __init__(self, all_obj, total):
        self.total_elements = total

        g00, g0, g1, g2, g3, g4, v1, v2 = tee(all_obj, 8)

        self.encoding = pref_encoding([obj.encoding for obj in g00])
        self.system = sum(obj.system for obj in g0)
        if total > 1:
            self.fieldAvgCount = statistics.mean(obj.count for obj in g3)
        else:
            self.fieldAvgCount = min((obj.count for obj in g3))

        self.valueUsedBytes = sum(obj.valueUsedBytes for obj in v1)
        self.valueAlignedBytes = sum(obj.valueAlignedBytes for obj in v2)

    def __enter__(self):
        return self

    def __exit__(self, *exc):
        return False


class List:
    def __init__(self, redis):
        """
        :param RmaRedis redis:
        :return:
        """
        self.redis = redis

    def analyze(self, keys):
        key_stat = {
            'headers': ['Match', "Count", "Avg Count", "Value mem", "Real", "Ratio", "System", "Encoding", "Total"],
            'data': []
        }
        # Undone Prefered encoding
        for pattern, data in keys.items():
            agg = ListAggreegator((ListStatEntry(x, self.redis) for x in data), len(data))

            stat_entry = [
                pattern,
                len(data),
                agg.fieldAvgCount,
                agg.valueUsedBytes,
                agg.valueAlignedBytes,
                agg.valueAlignedBytes / (agg.valueUsedBytes if agg.valueUsedBytes > 0 else 1),
                agg.system,
                agg.encoding,
                agg.valueAlignedBytes + agg.system
            ]

            key_stat['data'].append(stat_entry)

        key_stat['data'].sort(key=lambda x: x[8], reverse=True)
        key_stat['data'].append(make_total_row(key_stat['data'], ['Total:', sum, 0, sum, sum, 0, sum, '', sum]))

        return [
            "List stat",
            key_stat
        ]



import logging
from rma.redis import *
from tqdm import tqdm


class Scanner:
    """
    Get all keys from Redis database with given match and limits. If limit specified would be retrieved not more then
    limit keys.
    """
    logger = logging.getLogger(__name__)

    def __init__(self, redis, match="*", accepted_types=None):
        """
        :param RmaRedis redis:
        :param str match: Wild card match supported in Redis SCAN command
        :return:
        """
        self.keys = {
            REDIS_TYPE_ID_STRING: [],
            REDIS_TYPE_ID_HASH: [],
            REDIS_TYPE_ID_LIST: [],
            REDIS_TYPE_ID_SET: [],
            REDIS_TYPE_ID_ZSET: [],
            REDIS_TYPE_ID_UNKNOWN: [],
        }

        self.redis = redis
        self.match = match
        self.accepted_types = accepted_types[:] if accepted_types else REDIS_TYPE_ID_ALL
        self.resolve_types_script = self.redis.register_script("""
            local ret = {}
            for i = 1, #KEYS do
                ret[i] = redis.call("TYPE", KEYS[i])
            end
            return ret
        """)

    def __enter__(self):
        return self

    def __exit__(self, *exc):
        self.keys.clear()
        return False

    def batch_scan(self, count=1000, batch_size=3000):
        ret = []
        for key in self.redis.scan_iter(self.match, count=count):
            ret.append(key)
            if len(ret) == batch_size:
                yield from self.resolve_types(ret)

        if len(ret):
            yield from self.resolve_types(ret)

    def resolve_types(self, ret):
        key_with_types = self.resolve_types_script(ret)
        for i in range(0, len(ret)):
            yield key_with_types[i], ret[i]

        ret.clear()

    def scan(self, limit=1000):
        with tqdm(total=min(limit, self.redis.dbsize()), desc="Match {0}".format(self.match),
                  miniters=1000) as progress:

            total = 0

            for key_tuple in self.batch_scan():
                key_type, key_name = key_tuple
                if not key_name:
                    self.logger.warning(
                        '\r\nWarning! Scan iterator return key with empty name `` and type %s' % key_type)
                    continue

                to_id = redis_type_to_id(key_type)
                if to_id in self.accepted_types:
                    self.keys[to_id].append(key_name.decode("utf-8"))

                progress.update()

                total += 1
                if total > limit:
                    logging.info("\r\nLimit {0} reached".format(limit))
                    break

        return self.keys



#!/usr/bin/env python

from rma import __version__

long_description = '''
RMA is a console tool to scan Redis key space in real time and aggregate memory usage statistic by key patterns. You may
use this tools without maintenance on production servers. You can scanning by all or selected Redis types such as "string",
"hash", "list", "set", "zset" and use matching pattern as you like. RMA try to discern key names by patterns, for example
if you have keys like 'user:100' and 'user:101' application would pick out common pattern 'user:*' in output so you can
analyze most memory distressed data in your instance.
'''

sdict = {
    'name': 'rma',
    'version': __version__,
    'description': 'Utilities to profile Redis RAM usage',
    'long_description': long_description,
    'url': 'https://github.com/gamenet/redis-memory-analyzer',
    'author': 'Nikolay Bondarenko',
    'author_email': 'misterionkell@gmail.com',
    'maintainer': 'Nikolay Bondarenko',
    'maintainer_email': 'misterionkell@gmail.com',
    'keywords': ['Redis', 'Memory Profiler'],
    'license': 'MIT',
    'packages': ['rma', 'rma.helpers', 'rma.reporters', 'rma.rule', 'rma.cli'],
    'package_data': {'rma.cli': ['*.template']},
    'test_suite': 'tests.all_tests',
    'entry_points': {
        'console_scripts': [
            'rma = rma.cli.rma_cli:main',
        ],
    },
    'classifiers': [
        'Development Status :: 3 - Alpha',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python'
    ],
}

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

setup(**sdict, requires=['tqdm', 'tabulate', 'redis'])



from rma.helpers.formating import floored_percentage, pref_encoding, make_total_row
from rma.helpers.numbers import is_num, min_ge, next_power_of_2, is_power2

__all__ = ['floored_percentage', 'pref_encoding', 'make_total_row', 'is_num', 'min_ge', 'next_power_of_2', 'is_power2']
